/*  Simple template class for quaternions
 *  Copyright (C) R. Bourquin, 2008
 *  License: GNU GPL v2.0 or above
 */

#ifndef QUARealERNION_HH
#define QUARealERNION_HH
#include <cassert>
#include <cmath>

template <class Real>
class Quaternion {
    private:
        Real t; /**< First number of the quaternion */
        Real u; /**< Second number of the quaternion */
        Real v; /**< Realhird number of the quaternion */
        Real w; /**< Fourth number of the quaternion */
    public:
        Quaternion();
        Quaternion(Real a, Real b, Real c, Real d);

        Real get_t(){ return t; };
        Real get_u(){ return u; };
        Real get_v(){ return v; };
        Real get_w(){ return w; };

        Quaternion operator+(Quaternion q);
        Quaternion operator-(Quaternion q);
        Quaternion operator*(Quaternion q);
        Quaternion operator*(Real r);
};

/** Constructor with zero arguments. Do not use it.
 */
template <class Real>
Quaternion<Real>::Quaternion() {
    t = Real();
    u = Real();
    v = Real();
    w = Real();
};

/** Constructor for actually constructing a complex number.
 *  The two arguments are real numbers. They represent the real and
 *  the imaginary part of the complex number.
 */
template <class Real>
Quaternion<Real>::Quaternion(Real a, Real b, Real c, Real d) {
    t = a;
    u = b;
    v = c;
    w = d;
};

/** Define the addition of two quaternions.
 */
template <class Real>
Quaternion<Real> Quaternion<Real>::operator+(Quaternion q) {
    return Quaternion<Real>(t + q.get_t(), u + q.get_u(), v + q.get_v(), w + q.get_w());
};

/** Define the subtraction of two quaternions.
 */
template <class Real>
Quaternion<Real> Quaternion<Real>::operator-(Quaternion q) {
    return Quaternion<Real>(t - q.get_t(), u - q.get_u(), v - q.get_v(), w - q.get_w());
};

/** Define the product of two quaternions.
 */
template <class Real>
Quaternion<Real> Quaternion<Real>::operator*(Quaternion q) {
    Real t_temp = t*q.get_t()-u*q.get_u()-v*q.get_v()-w*q.get_w();
    Real u_temp = t*q.get_u()+u*q.get_t()+v*q.get_w()-w*q.get_v();
    Real v_temp = t*q.get_v()-u*q.get_w()+v*q.get_t()+w*q.get_u();
    Real w_temp = t*q.get_w()+u*q.get_v()-v*q.get_u()+w*q.get_t();
    return Quaternion<Real>(t_temp, u_temp, v_temp, w_temp);
};

/** Define the product of a quaternions and a real number.
 */
template <class Real>
Quaternion<Real> Quaternion<Real>::operator*(Real r) {
    return Quaternion<Real>(r*t, r*u, r*v, r*w);
};

// -------------- NON CLASS MEMBER FUNCTIONS --------------------- //

/** Define the product of a quaternions and a real number.
 */
template <class Real>
Quaternion<Real> operator*(Real r, Quaternion<Real> q) {
    return Quaternion<Real>(r*q.get_t(), r*q.get_u(), r*q.get_v(), r*q.get_w());
};

/** Returns conjugate of the quaternion.
 */
template <class Real>
Quaternion<Real> conjugate(Quaternion<Real> q) {
    return Quaternion<Real>(q.get_t(), -q.get_u(), -q.get_v(), -q.get_w());
};

/** Calculate and return the absolute value of the quaternion.
 *  It is defined as sqrt(q*conj(q)).
 */
template <class Real>
Real abs(Quaternion<Real> q) {
    return sqrt( (q*conjugate(q)).get_t() );
};

/** Returns the inverse of the quaternion.
 */
template <class Real>
Quaternion<Real> inverse(Quaternion<Real> q) {
    Real len = Real(1.0/(abs(q)*abs(q)));
    assert( len != Real(0) );
    return conjugate(q)*len;
};

/** Returns scalar of the quaternion
 */
template <class Real>
Quaternion<Real> scalar(Quaternion<Real> q) {
    return Real(1/2.0)*(q+conjugate(q));
};

/** Returns scalar of the quaternion
 */
template <class Real>
Quaternion<Real> vector(Quaternion<Real> q) {
    return Real(1/2.0)*(q-conjugate(q));
};

#endif
